"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Primitives = exports.Setup = void 0;
// External crypto
const circomlib_1 = __importStar(require("circomlib"));
const snarkjs_1 = require("snarkjs");
const websnark_1 = require("websnark");
const fixed_merkle_tree_1 = require("fixed-merkle-tree");
const utils_1 = require("websnark/src/utils");
// Some utils to work with hex numbers
const sdk_utils_1 = require("@tornado/sdk-utils");
// Parse some files
const sdk_data_1 = require("@tornado/sdk-data");
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ SETUP ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 * Several objects have to be set up, like the groth16 prover. All related types are
 * (will be) contained within this namespace.
 */
var Setup;
(function (Setup) {
    function getZeroElement() {
        return '21663839004416932945382355908790599225266501822907911457504978515578255421292';
    }
    Setup.getZeroElement = getZeroElement;
    function getDefaultHash() {
        // @ts-ignore
        return (left, right) => circomlib_1.mimcsponge.multiHash([(0, snarkjs_1.bigInt)(left), (0, snarkjs_1.bigInt)(right)]).toString();
    }
    Setup.getDefaultHash = getDefaultHash;
    function getProvingKey() {
        return sdk_data_1.Files.loadRawSync('circuits/tornadoProvingKey.bin').buffer;
    }
    Setup.getProvingKey = getProvingKey;
    function getTornadoCircuit() {
        return sdk_data_1.Json.loadSync('circuits/tornado.json');
    }
    Setup.getTornadoCircuit = getTornadoCircuit;
    let cachedGroth16Prover = null;
    async function getGroth16() {
        const defaultParams = { wasmInitialMemory: 5000 };
        if (!cachedGroth16Prover)
            cachedGroth16Prover = await (0, websnark_1.buildGroth16)(defaultParams);
        return cachedGroth16Prover;
    }
    Setup.getGroth16 = getGroth16;
    function terminateGroth16() {
        cachedGroth16Prover.terminate();
        cachedGroth16Prover = null;
    }
    Setup.terminateGroth16 = terminateGroth16;
})(Setup = exports.Setup || (exports.Setup = {}));
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Primitives ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
/**
 * These are the important cryptographic primitives Tornado Cash uses.
 */
var Primitives;
(function (Primitives) {
    function calcPedersenHash(pedersenHashData) {
        return circomlib_1.default.babyJub.unpackPoint(circomlib_1.default.pedersenHash.hash(pedersenHashData.msg))[0];
    }
    Primitives.calcPedersenHash = calcPedersenHash;
    function createNote(msg) {
        return sdk_utils_1.HexUtils.bufferToHex(msg, 62);
    }
    Primitives.createNote = createNote;
    function parseNote(hexNote) {
        const _hexNote = hexNote.split('_')[1] ?? hexNote;
        const buffer = Buffer.from(_hexNote.slice(2), 'hex');
        return createDeposit({
            // @ts-expect-error
            nullifier: snarkjs_1.bigInt.leBuff2int(buffer.subarray(0, 31)),
            // @ts-expect-error
            secret: snarkjs_1.bigInt.leBuff2int(buffer.subarray(31, 62))
        });
    }
    Primitives.parseNote = parseNote;
    function createDeposit(input = {
        nullifier: sdk_utils_1.NumberUtils.randomBigInteger(31),
        secret: sdk_utils_1.NumberUtils.randomBigInteger(31)
    }) {
        // @ts-expect-error
        let preimage = Buffer.concat([input.nullifier.leInt2Buff(31), input.secret.leInt2Buff(31)]);
        let commitment = calcPedersenHash({ msg: preimage });
        let commitmentHex = sdk_utils_1.HexUtils.bigIntToHex(commitment);
        // @ts-expect-error
        let nullifierHash = calcPedersenHash({ msg: input.nullifier.leInt2Buff(31) });
        let nullifierHex = sdk_utils_1.HexUtils.bigIntToHex(nullifierHash);
        return {
            nullifier: input.nullifier,
            secret: input.secret,
            preimage: preimage,
            commitment: commitment,
            hexCommitment: commitmentHex,
            nullifierHash: nullifierHash,
            hexNullifierHash: nullifierHex
        };
    }
    Primitives.createDeposit = createDeposit;
    function buildMerkleTree(inputs) {
        return new fixed_merkle_tree_1.MerkleTree(inputs.height, inputs.leaves, {
            hashFunction: Setup.getDefaultHash(),
            zeroElement: Setup.getZeroElement()
        });
    }
    Primitives.buildMerkleTree = buildMerkleTree;
    async function calcDepositProofs(inputs) {
        const proofs = [];
        const groth16 = await Setup.getGroth16();
        const circuit = Setup.getTornadoCircuit();
        const provingKey = Setup.getProvingKey();
        for (let i = 0, len = inputs.length; i < len; i++) {
            const input = inputs[i];
            // TODO: remove try and const again after fixing
            let proofData;
            // Compute Merkle Proof
            // The ts return is noted as `pathIndex` but using this we get an undefined because it is really `pathIndices`???
            // TODO: Bug that needs to be fixed (above)
            const { pathElements, pathIndices } = input.public.tree.path(input.public.leafIndex);
            proofs.push([]);
            try {
                proofData = await (0, utils_1.genWitnessAndProve)(groth16, {
                    // Public inputs
                    root: input.public.root,
                    // @ts-expect-error
                    nullifierHash: (0, snarkjs_1.bigInt)(input.public.hexNullifierHash),
                    // @ts-expect-error
                    recipient: (0, snarkjs_1.bigInt)(input.public.recipientAddress),
                    // @ts-expect-error
                    relayer: (0, snarkjs_1.bigInt)(input.public.relayerAddress),
                    // @ts-expect-error
                    fee: (0, snarkjs_1.bigInt)(input.public.fee),
                    //
                    refund: input.public.refund,
                    // Private inputs
                    nullifier: input.private.nullifier,
                    secret: input.private.secret,
                    pathElements: pathElements,
                    pathIndices: pathIndices
                }, circuit, provingKey);
            }
            catch (err) {
                groth16.terminate();
                throw sdk_utils_1.ErrorUtils.ensureError(err);
            }
            proofs[i].push((0, utils_1.toSolidityInput)(proofData).proof);
            proofs[i].push(String(input.public.root), input.public.hexNullifierHash, '0x' + sdk_utils_1.HexUtils.prepareAddress(input.public.recipientAddress, 20), '0x' + sdk_utils_1.HexUtils.prepareAddress(input.public.relayerAddress, 20), sdk_utils_1.HexUtils.bigIntToHex(input.public.fee), sdk_utils_1.HexUtils.bigIntToHex(input.public.refund));
        }
        // Done.
        Setup.terminateGroth16();
        return proofs;
    }
    Primitives.calcDepositProofs = calcDepositProofs;
})(Primitives = exports.Primitives || (exports.Primitives = {}));
//# sourceMappingURL=index.js.map